class Statement:
	def __init__(self, keyword, *args):
		self.keyword = keyword
		self.args = list(args)
		self.sub = []

	def append(self, v):
		self.args.append(v)

	def write_to_file(self, f, indent=0):
		s = self.keyword
		for a in self.args:
			if type(a)==float:
				s += " {:#.6g}".format(a)
			elif type(a)==bool:
				if a:
					s += " true"
				else:
					s += " false"
			elif type(a)==str:
				s += ' "{}"'.format(a)
			elif type(a)==Resource:
				s += ' "{}"'.format(a.name)
			else:
				s += " {}".format(a)

		tabs = '\t'*indent
		f.write("{}{}".format(tabs, s))
		if self.sub:
			f.write("\n{}{{\n".format(tabs))
			for s in self.sub:
				s.write_to_file(f, indent+1)
			f.write("{}}};\n".format(tabs))
		else:
			f.write(";\n")


class Token:
	def __init__(self, text):
		self.text = text

	def __str__(self):
		return self.text


class Resource:
	def __init__(self, name, keyword):
		self.name = name
		self.keyword = keyword
		self.statements = []
		self.references = []

	def create_reference_statement(self, keyword, *args):
		if len(args)<1:
			raise TypeError("create_reference_statement expected at least 2 arguments, got {}".format(1+len(args)))

		resources = []
		for a in args:
			if type(a)==Resource:
				resources.append(a)

		if not resources:
			raise TypeError("create_reference_statement expected a Resource argument, found none")

		self.references += [r for r in resources if r not in self.references]
		return Statement(keyword, *args)

	def create_embed_statement(self, keyword, *args):
		if len(args)<1:
			raise TypeError("create_embed_statement expected at least 2 arguments, got {}".format(1+len(args)))
		res = args[-1]
		if type(res)!=Resource:
			raise TypeError("create_embed_statement expected a Resource as last argument, got {}".format(type(args[-1]).__name__))

		self.references += [r for r in res.references if r not in self.references]
		st = Statement(keyword, *args[:-1])
		st.sub = res.statements
		return st

	def collect_references(self):
		refs = []
		for r in self.references:
			refs += [e for e in r.collect_references() if e not in refs]
		refs += [e for e in self.references if e not in refs]
		return refs

	def write_to_file(self, fn):
		with open(fn, "w") as out_file:
			for s in self.statements:
				s.write_to_file(out_file)

	def write_collection(self, fn, *, exclude_self=False, filter=None):
		items = self.collect_references()
		if not exclude_self:
			items.append(self)
		with open(fn, "w") as out_file:
			for i in items:
				if filter and not filter(i):
					continue
				st = Statement(i.keyword, i.name)
				st.sub = i.statements
				st.write_to_file(out_file)
